<?php

namespace Win7\Phone\DataTable;

use Win7\Application\Common\StringUtils;
use Win7\Application\Manager\DataTable\ColumnDefinition;
use Win7\Application\Manager\FilterParams;
use Win7\Application\Manager\TableManager;
use Win7\Phone\Entity\Call;
use Win7\Phone\Repository\CallRepository;

/**
 * @namespace Win7\Phone\DataTable
 * @author Thiago Daher
 *
 * @method CallRepository getRepository()
 */
class CallTable extends TableManager
{

    /**
     * @return string
     */
    protected function getRepositoryClass(): string
    {
        return CallRepository::class;
    }

    /**
     * @return array|\Win7\Application\Manager\DataTable\ColumnDefinition[]
     */
    protected function getColumnDefinitions(): array
    {
        return [
            'id' => new ColumnDefinition([
                'position'    => 0,
                'orderable'   => false,
                'searchable'  => false,
                'search_type' => ColumnDefinition::SEARCH_EQUAL,
            ]),

            'start_time' => new ColumnDefinition([
                'position'    => 1,
                'orderable'   => true,
                'searchable'  => true,
                'search_type' => ColumnDefinition::SEARCH_LIKE,
            ]),

            'source' => new ColumnDefinition([
                'position'     => 2,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setSource'],
            ]),

            'real_source' => new ColumnDefinition([
                'position'     => 3,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setRealSource'],
            ]),

            'destination' => new ColumnDefinition([
                'position'     => 4,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setDestination'],
            ]),

            'cost' => new ColumnDefinition([
                'position'     => 5,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setCost'],
            ]),

            'talking_duration' => new ColumnDefinition([
                'position'     => 6,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setTalkingDuration'],
            ]),

            'ringing_duration' => new ColumnDefinition([
                'position'     => 7,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setRingingDuration'],
            ]),

            'total_duration' => new ColumnDefinition([
                'position'     => 8,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setTotalDuration'],
            ]),

            'adicionado' => new ColumnDefinition([
                'position'     => 9,
                'orderable'    => false,
                'searchable'   => false,
                'load_from_db' => false,
                'callback'     => [$this, 'setAdicionado'],
            ]),

            'answered' => new ColumnDefinition([
                'position'    => 10,
                'orderable'   => true,
                'searchable'  => true,
                'search_type' => ColumnDefinition::SEARCH_LIKE,
            ]),

            'valido' => new ColumnDefinition([
                'position'     => 11,
                'orderable'    => false,
                'searchable'   => false,
                'search_type'  => ColumnDefinition::SEARCH_LIKE,
                'load_from_db' => false,
                'callback'     => [$this, 'setValido'],
            ]),
        ];
    }

    /**
     * @param array $context
     * @return string
     */
    public function setCost(array $context): string
    {
        $valores = $this->getParams()->getOption('valores');

        if ($context['charge_type'] === Call::NO_CHARGE || !is_array($valores) || count($valores) < 2) {
            return 'R$ 0,00';
        }

        if ($context['charge_type'] === Call::CHARGE_LANDLINE) {
            return StringUtils::formatMoney($context['charge_minutes'] * (float)$valores[1]);
        }

        return StringUtils::formatMoney($context['charge_minutes'] * (float)$valores[0]);
    }

    /**
     * @param array $context
     * @return bool
     * @throws \ReflectionException
     */
    public function setAdicionado(array $context): bool
    {
        return $this->getRepository()->findQuantidadeContatos(
                StringUtils::filter3cxNumber($context['source_caller_id'], $context['source_display_name']),
                $this->getParams()->getOption('cliente')
            ) > 0;
    }

    /**
     * @param array $context
     * @return string
     */
    public function setRealSource(array $context): string
    {
        return StringUtils::filter3cxNumber($context['source_caller_id'], $context['source_display_name']);
    }

    /**
     * @param array $context
     * @return bool
     */
    public function setValido(array $context): bool
    {
        $numero = StringUtils::filter3cxNumber($context['source_caller_id'], $context['source_display_name']);

        return $numero !== $context['source_caller_id'];
    }

    /**
     * @param array $context
     * @return string
     */
    public function setDestination(array $context): string
    {
        return StringUtils::filter3cxNumber($context['destination_caller_id'], $context['destination_caller_id']);
    }

    /**
     * @param array $context
     * @return string
     */
    public function setSource(array $context): string
    {
        $sourceRamal       = $context['source_ramal'];
        $destinationRamal  = $context['destination_ramal'];
        $sourceFilter      = strrev(preg_replace("/[^0-9]/", '', $context['source_filter']));
        $destinationFilter = strrev(preg_replace("/[^0-9]/", '', $context['destination_filter']));

        foreach ($this->getParams()->getOption('nomes') as $ramal => $nome) {
            if ($sourceRamal === $ramal ||
                $destinationRamal === $ramal ||
                str_starts_with($sourceFilter, $ramal) ||
                str_starts_with($destinationFilter, $ramal)
            ) {
                return $nome;
            }
        }

        return '';
    }

    /**
     * @param array $context
     * @return string
     */
    public function setTalkingDuration(array $context): string
    {
        return StringUtils::convertDurationSeconds($context['total_talking_duration']);
    }

    /**
     * @param array $context
     * @return string
     */
    public function setRingingDuration(array $context): string
    {
        return StringUtils::convertDurationSeconds($context['total_ringing_duration']);
    }

    /**
     * @param array $context
     * @return string
     */
    public function setTotalDuration(array $context): string
    {
        return StringUtils::convertDurationSeconds($context['total_talking_duration'] + $context['total_ringing_duration']);
    }

    /**
     * @param \Win7\Application\Manager\FilterParams $params
     * @return array
     * @throws \ReflectionException
     * @throws \Win7\Application\Exception\UserException
     */
    public function getTableList(FilterParams $params): array
    {
        $result    = parent::getTableList($params);
        $valores   = $this->getParams()->getOption('valores');
        $showTotal = $params->getOption('carregarTotais') == 1;

        if ($showTotal && is_array($valores) && count($valores) === 2) {
            $total = $this->getRepository()->countChargedCalls($params);
            $value = $total->getCellphoneMinutes() * (float)$valores[0];
            $value += $total->getLandlineMinutes() * (float)$valores[1];

            $result['valor'] = StringUtils::formatMoney($value);
        } else {
            $result['valor'] = 'R$ 0,00';
        }

        $result['permitirLeads'] = $this->getParams()->getOption('permitirLeads');
        $result['urlContato']    = $this->getParams()->getOption('urlContato');

        return $result;
    }
}