<?php

namespace Win7\Phone\Manager;

use Win7\Application\Common\StringUtils;
use Win7\Application\Manager\FilterParams;
use Win7\Application\Manager\Manager;
use Win7\Application\Repository\RepositoryManager;
use Win7\Phone\DataTable\CallTable;
use Win7\Phone\Repository\CallRepository;

/**
 * @namespace Win7\Phone\Manager
 * @author Thiago Daher
 *
 * @method CallRepository getRepository()
 */
class CallManager extends Manager
{

    /**
     * @var \Win7\Phone\DataTable\CallTable
     */
    private CallTable $callTable;

    /**
     * @param \Win7\Application\Repository\RepositoryManager $repositoryManager
     * @param \Win7\Phone\DataTable\CallTable $callTable
     */
    public function __construct(RepositoryManager $repositoryManager, CallTable $callTable)
    {
        parent::__construct($repositoryManager);
        $this->callTable = $callTable;
    }

    /**
     * @return string
     */
    protected function getRepositoryClass(): string
    {
        return CallRepository::class;
    }

    /**
     * Obtém os dados de segmento da ligação.
     *
     * @param mixed $id representa o id da ligação
     * @return array|null
     * @throws \ReflectionException
     */
    public function getDadosLigacao($id): ?array
    {
        $ligacao = $this->getRepository()->findById($id);

        if (!$ligacao) {
            return null;
        }

        $resultado = [];

        foreach ($ligacao->segments as $segment) {
            $segmentData                     = $segment->toArray();
            $segmentData['ringing_duration'] = StringUtils::convertDurationSeconds($segmentData['ringing_duration']);
            $segmentData['talking_duration'] = StringUtils::convertDurationSeconds($segmentData['talking_duration']);
            $segmentData['answered']         = $segmentData['answered'] ? 't' : 'n';
            $resultado[]                     = $segmentData;
        }

        return $resultado;
    }

    /**
     * @param array $params
     * @return int
     * @throws \ReflectionException
     */
    public function contarLigacoes(array $params): int
    {
        $params = new FilterParams($params);

        if ($params->getOption('esconderEfetuadas')) {
            $params->setOption('efetuadas', false);
            $params->setOption('recebidas', true);
        }

        return $this->getRepository()->countCalls($params);
    }

    /**
     * Adaptação do método antigo para obter todas as ligações.
     * Ainda precisa ser muito melhorado, mas foi deixado assim para fins de compatibilidade.
     *
     * @param array $params
     * @return array
     * @throws \ReflectionException
     * @throws \Win7\Application\Exception\UserException
     */
    public function obterTodasLigacoes(array $params): array
    {
        $params       = new FilterParams($params);
        $tableResult  = $this->callTable->getAllData($params);
        $result       = [];
        $ramais       = [];
        $naoAtendidas = 0;

        foreach ($tableResult['data'] as $data) {
            if (!in_array($data[2], $ramais)) {
                $ramais[] = $data[2];
            }

            $result[] = [
                'fonte'        => $data[2],
                'origem'       => $data[3],
                'destino'      => $data[4],
                'data'         => $params->getOption('grafico') ? date('Y-m-d', strtotime($data[1])) : $data[1],
                'valor'        => $data[5],
                'duracao'      => $data[6],
                'valido'       => $data[11] ? '1' : '0',
                'id'           => $data[0],
                'atendida'     => $data[10] ? 'Atendida' : 'Não Atendida',
                'espera'       => $data[7],
                'duracaofinal' => $data[8],
            ];

            if (!$data[10]) {
                $naoAtendidas++;
            }
        }

        $extra = [];

        if (count($result) > 0) {
            $extra['data2']  = date('Y-m-d', strtotime($result[0]['data']));
            $extra['data'] = date('Y-m-d', strtotime($result[count($result) - 1]['data']));
        }

        $extra['ids']           = $ramais;
        $extra['totalgasto']    = $tableResult['valor'];
        $extra['totalligacoes'] = $tableResult['recordsTotal'];
        $extra['naoatendidas']  = $naoAtendidas;

        array_unshift($result, $extra);

        return $result;
    }
}