<?php

namespace Win7\Phone\Repository;

use Closure;
use Illuminate\Database\Eloquent\Builder;
use Win7\Application\Manager\DataTable\ColumnDefinition;
use Win7\Application\Manager\FilterParams;
use Win7\Application\Repository\Repository;
use Win7\Phone\Domain\CallCostCountResult;
use Win7\Phone\Entity\Call;

/**
 * @namespace Win7\Phone\Repository
 * @author Thiago Daher
 *
 * @method Call|null findById($id)
 */
class CallRepository extends Repository
{

    /**
     * @var \Win7\Application\Manager\FilterParams
     */
    private FilterParams $params;

    /**
     * @return string
     */
    public function getEntityClass(): string
    {
        return Call::class;
    }

    /**
     * @return \Win7\Application\Manager\FilterParams
     */
    public function getParams(): FilterParams
    {
        return $this->params;
    }

    /**
     * @param \Win7\Application\Manager\FilterParams $params
     */
    public function setParams(FilterParams $params): void
    {
        $this->params = $params;
    }

    /**
     * @param \Win7\Application\Manager\FilterParams $params
     * @param bool $select
     * @return \Illuminate\Database\Eloquent\Builder
     */
    protected function getTableQuery(FilterParams $params, bool $select = true): Builder
    {
        $this->setParams($params);
        $query = parent::getTableQuery($params, $select);

        if ($select) {
            $this->doTableJoin($query, 'segments.id', ColumnDefinition::INNER_JOIN);
            $query->addSelect('call.source_caller_id as source_caller_id');
            $query->addSelect('call.source_display_name as source_display_name');
            $query->addSelect('call.destination_caller_id as destination_caller_id');
            $query->addSelect('call.destination_display_name as destination_display_name');
            $query->addSelect('call.charge_type as charge_type');
            $query->addSelect('call.charge_minutes as charge_minutes');
            $query->addSelect('call.total_ringing_duration as total_ringing_duration');
            $query->addSelect('call.total_talking_duration as total_talking_duration');
            $query->addSelect('segments.source_ramal as source_ramal');
            $query->addSelect('segments.destination_ramal as destination_ramal');
            $query->addSelect('segments.source_filter as source_filter');
            $query->addSelect('segments.destination_filter as destination_filter');
        }

        if (!$params->isSearch()) {
            $query->orderBy('segments.start_time', 'desc');
        }

        $query->where('segments.start_time', '>=', $params->getOption('dataInicial'));
        $query->where('segments.start_time', '<=', $params->getOption('dataFinal'));
        $query->where(Closure::fromCallable([$this, 'filtrarRamais']));

        return $query;
    }

    /**
     * @param \Illuminate\Database\Eloquent\Builder|\Illuminate\Database\Query\Builder $query
     * @return void
     */
    private function addCustomFilters($query)
    {
        $source      = $this->getParams()->getColumn('source');
        $realSource  = $this->getParams()->getColumn('real_source');
        $destination = $this->getParams()->getColumn('destination');

        if (!empty($realSource->getSearch())) {
            $search = strrev(preg_replace('/[^0-9]/', '', $realSource->getSearch()));
            $query->where('segments.source_filter', 'like', $search . '%');
        }

        if (!empty($source->getSearch())) {
            foreach ($this->getParams()->getOption('nomes') as $ramal => $nome) {
                if ($nome === $source->getSearch()) {
                    $query->where(Closure::fromCallable(function ($query) use ($ramal) {
                        $this->filtrarRamal($query, $ramal);
                    }));
                }
            }
        }

        if (!empty($destination->getSearch())) {
            $search = strrev(preg_replace('/[^0-9]/', '', $destination->getSearch()));
            $query->where('segments.destination_filter', 'like', $search . '%');
        }
    }

    /**
     * @param string $numero
     * @param string $idCliente
     * @return int
     */
    public function findQuantidadeContatos(string $numero, string $idCliente): int
    {
        $query = $this->getConnection()->query();
        $query->select('NU_TELEFONE1');
        $query->where('ID_CLIENTE_FK', '=', $idCliente);
        $query->where('NU_TELEFONE1', '=', $numero);
        $query->where('TXT_LEADTIPO', '=', 'Win7 Fone');
        $query->from('TBL_LEADS');

        return $query->count();
    }

    /**
     * @param \Win7\Application\Manager\FilterParams $params
     * @return \Win7\Phone\Domain\CallCostCountResult
     */
    public function countChargedCalls(FilterParams $params): CallCostCountResult
    {
        $cellPhoneQuery = $this->getCountCallsQuery($params);
        $landlineQuery  = clone $cellPhoneQuery;
        $cellPhoneQuery->where('call.charge_type', '=', Call::CHARGE_CELL_PHONE);
        $landlineQuery->where('call.charge_type', '=', Call::CHARGE_LANDLINE);

        $sql = sprintf(
            "SELECT %s, %s FROM (%s) as dados",
            'COUNT(id) as quantity',
            'SUM(charge_minutes) as minutes',
            $cellPhoneQuery->toSql()
        );

        $cellPhoneResult = $this->getConnection()->select($sql, $cellPhoneQuery->getBindings());
        $cellPhoneResult = !empty($cellPhoneResult) ? $cellPhoneResult[0] : [];

        $sql = sprintf(
            "SELECT %s, %s FROM (%s) as dados",
            'COUNT(id) as quantity',
            'SUM(charge_minutes) as minutes',
            $landlineQuery->toSql()
        );

        $landlineResult = $this->getConnection()->select($sql, $landlineQuery->getBindings());
        $landlineResult = !empty($landlineResult) ? $landlineResult[0] : [];

        $result = new CallCostCountResult();
        $result->setCellphoneCount($cellPhoneResult->quantity ?? 0);
        $result->setCellphoneMinutes($cellPhoneResult->minutes ?? 0);
        $result->setLandlineCount($landlineResult->quantity ?? 0);
        $result->setLandlineMinutes($landlineResult->minutes ?? 0);

        return $result;
    }

    /**
     * @param \Win7\Application\Manager\FilterParams $params
     * @return int
     */
    public function countCalls(FilterParams $params): int
    {
        $query = $this->getCountCallsQuery($params);

        if ($params->getOption('atendida') !== null) {
            $query->where('call.answered', '=', $params->getOption('atendida') ? '1' : '0');
        }

        $sql = sprintf(
            "SELECT %s FROM (%s) as dados",
            'COUNT(id) as quantity',
            $query->toSql()
        );

        $result = $this->getConnection()->select($sql, $query->getBindings());
        $result = !empty($result) ? $result[0] : [];

        return $result->quantity ?? 0;
    }

    /**
     * @param \Win7\Application\Manager\FilterParams $params
     * @return \Illuminate\Database\Eloquent\Builder
     */
    private function getCountCallsQuery(FilterParams $params): Builder
    {
        $this->setParams($params);
        $query = $this->newQuery();
        $query->addSelect('call.id');
        $query->addSelect('call.charge_minutes');
        $query->addSelect('call.charge_type');
        $query->join('call_segment as segments', 'call.id', '=', 'segments.call_id');
        $query->where('segments.start_time', '>=', $params->getOption('dataInicial'));
        $query->where('segments.start_time', '<=', $params->getOption('dataFinal'));
        $query->where(Closure::fromCallable([$this, 'filtrarRamais']));

        return $query;
    }

    /**
     * @param string $startDate
     * @param string $endDate
     * @return \Illuminate\Database\Eloquent\Builder[]|\Illuminate\Database\Eloquent\Collection|\Win7\Phone\Entity\Call[]
     */
    public function findCallsByPeriod(string $startDate, string $endDate)
    {
        $query = $this->newQuery();
        $query->where('call.start_time', '>=', $startDate);
        $query->where('call.start_time', '<=', $endDate);

        /** @var Call[]|\Illuminate\Database\Eloquent\Builder[]|\Illuminate\Database\Eloquent\Collection $result */
        $result = $query->get();

        return $result;
    }

    /**
     * @param \Win7\Application\Manager\FilterParams $params
     * @return int
     */
    public function getTableCount(FilterParams $params): int
    {
        $query = $this->getConnection()->query();
        $query->from('call_segment', 'segments');
        $query->selectRaw('count(distinct call_id) as quantidade');
        $query->where('segments.start_time', '>=', $params->getOption('dataInicial'));
        $query->where('segments.start_time', '<=', $params->getOption('dataFinal'));
        $query->where(Closure::fromCallable([$this, 'filtrarRamais']));

        return $query->value('quantidade');
    }


    /**
     * @param \Illuminate\Database\Eloquent\Builder|\Illuminate\Database\Query\Builder $query
     * @return void
     */
    public function filtrarRamais($query)
    {
        if(empty($this->getParams()->getOption('ramais'))){
            $query->where('segments.source_ramal', '=', '-1');
            return;
        }
        
        $ramaisResultantes = [];

        foreach ($this->getParams()->getOption('ramais') as $ramal) {
            $ramal               = $this->filtrarRamal($query, $ramal);
            $ramaisResultantes[] = $ramal;
        }

        $this->getParams()->setOption('ramais', $ramaisResultantes);
    }

    /**
     * @param \Illuminate\Database\Eloquent\Builder|\Illuminate\Database\Query\Builder $query
     * @param $ramal
     * @return string
     */
    private function filtrarRamal($query, $ramal): string
    {
        $made     = $this->getParams()->getOption('efetuadas');
        $received = $this->getParams()->getOption('recebidas');

        if (strlen($ramal) <= 5) {
            if ($made === null || $made === true) {
                $query->orWhere('segments.source_ramal', '=', $ramal);
            }
            if ($received === null || $received === true) {
                $query->orWhere('segments.destination_ramal', '=', $ramal);
            }

            return $ramal;
        }

        $ramal = strrev(preg_replace("/[^0-9]/", '', $ramal));

        if ($made === null || $made === true) {
            $query->orWhere('segments.source_filter', 'like', $ramal . '%');
        }

        if ($received === null || $received === true) {
            $query->orWhere('segments.destination_filter', 'like', $ramal . '%');
        }

        return $ramal;
    }
}